@@ -20,8 +20,9 @@ import traceback
 import weakref
 import functools
 import operator
+import warnings
 
-__version__ = "0.4.1"  # automatically patched by setup.py when packaging
+__version__ = "0.4.2"  # automatically patched by setup.py when packaging
 
 # from six.py's strategy
 INTEGER_TYPES = None
@@ -54,6 +55,7 @@ except ImportError:  # py2 has no enum
 
 if sys.version_info[0] == 2:
     from socket import error as ConnectionError  # ConnectionError not defined in python2, this is next closest thing
+    from socket import error as ConnectionResetError  # as above
 
 
 class ThreadingTCPServer(socketserver.ThreadingMixIn, socketserver.TCPServer):
@@ -378,7 +380,7 @@ class BridgeCommandHandler(socketserver.BaseRequestHandler):
             # only get here if the client has requested we shutdown the bridge
             self.server.bridge.logger.debug("Receiver thread exited - bridge shutdown requested")
             self.server.bridge.shutdown()
-        except BridgeClosedException:
+        except (BridgeClosedException, ConnectionResetError):
             pass  # expected - the client has closed the connection
         except Exception as e:
             # something weird went wrong?
@@ -493,7 +495,7 @@ class BridgeConn(object):
         """ Set up the bridge connection - only instantiates a connection as needed """
         self.host = connect_to_host
         self.port = connect_to_port
-
+    
         # get a reference to the bridge's logger for the connection
         self.logger = bridge.logger
 
@@ -513,6 +515,10 @@ class BridgeConn(object):
         # we'll keep all the types forever (including handles to bridgedcallables in them) because types are super-likely
         # to be reused regularly, and we don't want to keep deleting them and then having to recreate them all the time.
         self.cached_bridge_types = dict()
+        
+        # if the bridge has requested a local_call_hook/local_eval_hook, record that
+        self.local_call_hook = bridge.local_call_hook
+        self.local_eval_hook = bridge.local_eval_hook
 
     def __del__(self):
         """ On teardown, make sure we close our socket to the remote bridge """
@@ -742,7 +748,7 @@ class BridgeConn(object):
             result = setattr(target, name, value)
         except Exception as e:
             result = e
-            traceback.print_exc()
+            traceback.print_exc() # TODO - this and other tracebacks, log with info about what's happening
 
         return result
 
@@ -778,7 +784,11 @@ class BridgeConn(object):
         result = None
         try:
             target_callable = self.get_object_by_handle(handle)
-            result = target_callable(*args, **kwargs)
+            # call the target function, or the hook if we've registered one
+            if self.local_call_hook is None:
+                result = target_callable(*args, **kwargs)
+            else:
+                result = self.local_call_hook(self, target_callable, *args, **kwargs)
         except EXCEPTION_TYPES as e:
             result = e
             if not isinstance(e, Exception):
@@ -975,7 +985,14 @@ class BridgeConn(object):
             """ the import __main__ trick allows accessing all the variables that the bridge imports, 
             so evals will run within the global context of what started the bridge, and the arguments 
             supplied as kwargs will override that """
-            result = eval(args[EXPR], importlib.import_module('__main__').__dict__, args[KWARGS])
+            eval_expr = args[EXPR]
+            eval_globals = importlib.import_module('__main__').__dict__
+            eval_locals = args[KWARGS]
+            # do the eval, or defer to the hook if we've registered one
+            if self.local_eval_hook is None:
+                result = eval(eval_expr, eval_globals, eval_locals)
+            else:
+                result = self.local_eval_hook(self, eval_expr, eval_globals, eval_locals)
             self.logger.debug("local_eval: Finished evaluating")
         except Exception as e:
             result = e
@@ -1091,14 +1108,16 @@ class BridgeConn(object):
         return bridge_type(self, obj_dict)
 
 
-class BridgeServer(threading.Thread):
+class BridgeServer(threading.Thread): # TODO - have BridgeServer and BridgeClient share a class
     """ Python2Python RPC bridge server 
 
         Like a thread, so call run() to run directly, or start() to run on a background thread
     """
     is_serving = False
+    local_call_hook = None
+    local_eval_hook = None
 
-    def __init__(self, server_host=DEFAULT_HOST, server_port=0, loglevel=None, response_timeout=DEFAULT_RESPONSE_TIMEOUT):
+    def __init__(self, server_host=DEFAULT_HOST, server_port=0, loglevel=None, response_timeout=DEFAULT_RESPONSE_TIMEOUT, local_call_hook=None, local_eval_hook=None):
         """ Set up the bridge.
 
             server_host/port: host/port to listen on to serve requests. If not specified, defaults to 127.0.0.1:0 (random port - use get_server_info() to find out where it's serving)
@@ -1127,6 +1146,12 @@ class BridgeServer(threading.Thread):
 
         # if we're starting the server, we need to make sure the flag is set to false
         GLOBAL_BRIDGE_SHUTDOWN = False
+        
+        # specify a callable to local_call_hook(bridge_conn, target_callable, *args, **kwargs) or
+        # local_eval_hook(bridge_conn, eval_expression, eval_globals_dict, eval_locals_dict) to
+        # hook local_call/local_eval to allow inspection/modification of calls/evals (e.g., forcing them onto a particular thread)
+        self.local_call_hook = local_call_hook
+        self.local_eval_hook = local_eval_hook
 
     def get_server_info(self):
         """ return where the server is serving on """
@@ -1149,8 +1174,12 @@ class BridgeServer(threading.Thread):
             self.server.server_close()
 
 
-class BridgeClient(object):
+class BridgeClient(object): # TODO make the external bridges inherit from this.
     """ Python2Python RPC bridge client """
+    
+    local_call_hook = None
+    local_eval_hook = None
+    _bridge = None
 
     def __init__(self, connect_to_host=DEFAULT_HOST, connect_to_port=DEFAULT_SERVER_PORT, loglevel=None, response_timeout=DEFAULT_RESPONSE_TIMEOUT, hook_import=False):
         """ Set up the bridge client
@@ -1176,6 +1205,17 @@ class BridgeClient(object):
             # We add it at the end, so we only catch imports that no one else wants to handle
             sys.path.append(repr(self.client))
             # TODO make sure we remove the finder when the client is torn down?
+        
+        self._bridge = self
+        
+    @property
+    def bridge(self):
+        """ for backwards compatibility with old examples using external_bridge.bridge.remote_import/etc,
+            before the external bridges just inherited from BridgeClient
+            Allow access, but warn about it
+        """
+        warnings.warn("Using <external_bridge>.bridge to get to remote_import/eval/shutdown is deprecated - just do <external_bridge>.remote_import/etc.", DeprecationWarning)
+        return self._bridge
 
     def remote_import(self, module_name):
         return self.client.remote_import(module_name)
